import copy
import logging
from abc import ABCMeta, abstractmethod
from queue import Queue
from threading import Lock
from typing import (
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Protocol,
    runtime_checkable,
)

from pydantic import BaseModel, Field

from logis.biz.sim import ISimProxy
from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG, Point, get_time, get_time_3d

from .model import TransportProperties


class ITransport(ISimProxy):
    """
    搬运设备，同时具备移动属性
    """

    def is_valid(self) -> bool:
        return True

    @classmethod
    def with_properties(cls, p: TransportProperties, **kwargs):
        inst = cls(**kwargs)
        inst.props = p
        return inst

    props: Optional[TransportProperties] = None

    def load(self, *args, **kwargs):
        """
        装载过程
        """
        t = get_time(self.props.load_distance, self.props.load_speed)
        yield self.env.timeout(t)

    def unload(self, *args, **kwargs):
        """
        卸载过程
        """
        l = self.props.unload_distance or self.props.load_distance
        v = self.props.unload_speed or self.props.load_speed
        t = get_time(l, v)
        yield self.env.timeout(t)

    def move(self, *args, **kwargs):
        """
        移动过程
        """
        d = self.props.distance_vector
        assert d is not None, "coordinates given can't be None"
        yield self.env.timeout(get_time_3d(d, self.props.speed))
        self.props.current_location = copy.copy(self.props.target_location)

    def find_path(self, *args, **kwargs):
        """
        寻找路径
        """
        yield self.env.timeout(0)


class ObstacleDetectorConfig(BaseModel):
    """
    障碍检测配置
    """

    agent_as_obstacle: bool = False
    rack_as_obstacle: bool = False
    end_point_as_obstacle: bool = False


class ObstacleDetectorOutput(BaseModel):

    obstacles: List[Point] = Field(default_factory=list)

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG


@runtime_checkable
class ObstacleDetector(Protocol):
    """
    障碍检测器
    """

    def detect_obstacles(
        self, config: Optional[ObstacleDetectorConfig] = None, *args, **kwargs
    ) -> ObstacleDetectorOutput:
        pass


class ManagedMovable(metaclass=ABCMeta):
    """
    按照外部manager规划路线进行移动
    """

    SIGNAL_RESTORE = "restore"

    @abstractmethod
    def get_target_location(self) -> Point:
        pass

    @abstractmethod
    def get_current_location(self) -> Point:
        pass

    @abstractmethod
    def _managed_move_(
        self, end: Point, start: Optional[Point] = None, **kwargs
    ) -> Generator:
        pass

    def __init__(self, *args, **kwargs):
        self.__control_queue__ = Queue()
        self.routine: Queue[Point] = Queue()
        self.__state__: Optional[Literal["paused", "stopped", "active"]] = None

        self.__last_routine__: List[Point] = []
        self.__lock__ = Lock()

    def clear_routine(self):
        """
        清除阻塞队列中的所有点
        """
        with self.__lock__:
            logging.debug("clear routine, before size is %s", self.routine.qsize())
            return self.routine.queue.clear()

    def update_routine(self, routine: Iterable[Point]):
        """
        更新阻塞队列中的路线:删除原有点并添加新的点
        """
        routine = list(routine)
        logging.debug("%s -> %s", self.__last_routine__, routine)
        self.clear_routine()
        for point in routine:
            self.routine.put(point)
        self.__last_routine__ = list(routine)

    def stop_moving_by_routine(self):
        """
        停止移动
        """
        self.__state__ = "stopped"
        logging.debug("%s is stopped", self)

    def pause_moving_by_routine(self):
        self.__state__ = "paused"
        logging.debug("%s is paused", self)
        self.__control_queue__.queue.clear()

    def restore_moving_by_routine(self):
        """
        恢复暂停的进度
        """
        if not self.is_paused:
            logging.warning("%s is not paused, can't restore", self)
            return
        self.__state__ = "active"
        self.__control_queue__.put(self.SIGNAL_RESTORE)
        logging.debug("%s is restored", self)

    def is_routine_empty(self) -> bool:
        return self.routine.empty()

    def wait_routine_empty(self, interval: float = 1):
        """
        等待移动完成
        """
        while not self.is_routine_empty():
            logging.debug(
                "%s is waiting for routine is empty, now size is %s",
                self.__class__.__name__,
                self.routine.qsize(),
            )
            yield from self.wait(interval)

    @abstractmethod
    def wait(self, time: float = 0):
        pass

    @abstractmethod
    def process(self, generator: Generator):
        pass

    @property
    def is_paused(self) -> bool:
        return self.__state__ == "paused"

    @property
    def is_stopped(self) -> bool:
        return self.__state__ == "stopped"

    @property
    def is_active(self) -> bool:
        return self.__state__ == "active"

    def _get_next_point(self, **kwargs) -> Point:
        """
        获取下一个点
        """
        if self.is_paused:
            return self.__control_queue__.get(**kwargs)
        return self.routine.get(**kwargs)

    def start_moving_by_routine(
        self,
        until_target: bool = False,
        until_empty: bool = False,
        wait_interval: float = 3,
        **kwargs,
    ):
        """
        按照阻塞队列中的路线移动
        """
        assert self.is_active is not True, "already moving by routine"
        logging.debug("%s start moving by routine", self)

        self.__state__ = "active"
        # TODO: 加锁?
        # 调用此方法即是强制开启移动，是否正在移动中由调用方判断
        last_point = None
        # 在运行过程中外部调用可能停止此过程
        while not (self.is_stopped):
            if until_empty and self.routine.empty():
                logging.debug("%s routine is empty, will stop", self)
                break
            try:
                next_point = self._get_next_point()
                if next_point == self.SIGNAL_RESTORE:
                    continue
            except:
                logging.debug("%s is waiting for next point", self)
                yield from self.wait(wait_interval)
                continue
            if self.is_stopped:
                break
            logging.debug(
                "%s last is %s,current is %s,next is %s",
                self,
                last_point,
                self.get_current_location(),
                next_point,
            )
            if last_point is not None and last_point != self.get_current_location():
                logging.warning(
                    "%s!=%s,who update current_location?",
                    last_point,
                    self.get_current_location(),
                )
            yield self.process(self._managed_move_(end=next_point, **kwargs))
            if (
                until_target
                and self.get_current_location() == self.get_target_location()
            ):
                break
            last_point = next_point

        self.stop_moving_by_routine()
        logging.info("%s stop moving by routine", self)


class AutonomousMovable(metaclass=ABCMeta):
    """
    自主移动
    """



import asyncio
import queue

from logis.iface import IQueue


def test_queue_type():

    assert asyncio.Queue is not queue.Queue
    assert asyncio.PriorityQueue is not queue.PriorityQueue

    assert issubclass(asyncio.Queue, IQueue)
    assert issubclass(asyncio.PriorityQueue, IQueue)

    assert issubclass(queue.Queue, IQueue)
    assert issubclass(queue.PriorityQueue, IQueue)

    q = queue.Queue()
    q.empty()



from pathlib import Path
from typing import Union


def resolve_model_name(model_path: Union[str, Path]):
    if not model_path:
        return None
    model_path = Path(model_path)
    return model_path.name


def resolve_excel_path(model_path: Union[str, Path]):
    if not model_path:
        return None
    model_path = Path(model_path)
    return model_path.joinpath(resolve_model_name(model_path) + ".xlsx")


def resolve_json_path(model_path: Union[str, Path]):
    if not model_path:
        return None
    model_path = Path(model_path)
    return model_path.joinpath("simulation_data.json")


def resolve_result_path(model_path: Union[str, Path]):
    if not model_path:
        return None
    model_path = Path(model_path)
    return model_path.joinpath("result.json")


def sim_data_dir(sim_id: str, data_dir: Path):
    if not sim_id:
        return data_dir
    return data_dir.joinpath(sim_id)



from typing import Generic, Iterable, Iterator, Optional, TypeAlias, TypeVar, Union

from pydantic import BaseModel

from logis.biz.sim import AgentId
from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG

from ..iface import AgentClass

TaskType = TypeVar("TaskType")


class AgentManifest(BaseModel, Generic[AgentClass, TaskType]):
    agent: Optional[AgentClass] = None
    agent_id: Optional[AgentId] = None

    tasks: Union[Iterable[TaskType], Iterator[TaskType]] = []

    __task_iter__: Optional[Iterator[TaskType]] = None

    def get_task_iter(self, is_new: bool = False):
        if is_new or self.__task_iter__ is None:
            self.__task_iter__ = iter(self.tasks)
        return self.__task_iter__

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG



import logging
import time
from pathlib import Path
from typing import List, Optional, Union

import psutil
from ipa.decorator import deprecated


def get_disk_usage(path: Union[str, Path], follow_symlink: bool = True):
    """
    获取指定路径的磁盘使用情况。

    TODO: 增加缓存；迭代效率

    Args:
        path (Union[str, Path]): 要获取磁盘使用情况的路径。
        follow_symlink (bool, optional): 是否跟随符号链接。默认值为 True。

    Returns:
        int: 路径占用的磁盘空间大小（字节）。
    """
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"path {path} not exists")

    if path.is_symlink() and follow_symlink:
        path = path.resolve()
    if path.is_file():
        byte_size = path.stat().st_size
    elif path.is_dir():
        byte_size = sum(get_disk_usage(f, follow_symlink) for f in path.glob("**/*"))
    else:
        raise ValueError(f"path {path} is not a file or dir")
    return byte_size


@deprecated("请使用 find_all_process_on_port 替代")
def find_process_on_port(port: int):
    """
    查找占用指定端口的进程，如果有多个只返回第一个。
    """
    for proc in psutil.process_iter(["pid", "name", "cmdline"]):
        if not proc.pid:
            continue
        connections = proc.net_connections(kind="inet")
        for conn in connections:
            if not conn.laddr:
                continue
            if conn.laddr.port == int(port):
                return proc
    return None


@deprecated("请使用 kill_all_process_on_port 替代")
def kill_process_on_port(port, max_try: Optional[int] = 1) -> Optional[int]:
    """
    强制释放占用指定端口的进程，如果有多个进程只处理第一个。

    Args:
        port (int): 要释放的端口号
        max_try (int, optional): 最大尝试次数。值为 None，代表无限次尝试。

    Returns:
        int | None: 成功释放进程的 PID，如果不存在对应进程则返回 None。
    """
    pid: Optional[int] = None
    try_count = 0
    while max_try is None or try_count < max_try:
        p = find_process_on_port(port)
        if not p or not p.pid:
            return pid
        pid = p.pid
        try:
            p.kill()
            logging.info("process on port %s been killed: %s", port, pid)
            return pid
        except Exception as e:
            try_count += 1
            logging.warning(
                "kill process on port %s failed(x%s): %s", port, try_count, e
            )
    raise RuntimeError(f"kill process on port {port} failed after {try_count} times")


def find_all_process_on_port(port: int) -> List[psutil.Process]:
    """
    查找占用指定端口的所有进程。
    """
    procs = []
    for proc in psutil.process_iter(["pid", "name", "cmdline"]):
        if not proc.pid:
            continue
        connections = proc.net_connections(kind="inet")
        for conn in connections:
            if not conn.laddr:
                continue
            if conn.laddr.port == int(port):
                procs.append(proc)
    return procs


def wait_port_idle(
    port: int, check_interval: float = 1.0, timeout: Optional[float] = None
):
    """
    等待指定端口空闲，即没有进程占用该端口。

    Args:
        port (int): 要等待的端口号
        check_interval (float, optional): 检查间隔，单位为秒。默认值为 1.0 秒。
        timeout (float, optional): 最大等待时间，单位为秒。默认值为 None，代表无限等待。

    Returns:
        bool: 如果端口空闲则返回 True，否则返回 False。
    """
    start = time.time()
    while True:
        if timeout is not None and time.time() - start > timeout:
            return False
        ps = find_all_process_on_port(port)
        if not ps:
            return True
        logging.info("there are %s process on port %s: %s", len(ps), port, ps)
        time.sleep(check_interval)


def kill_all_process_on_port(port, max_try: Optional[int] = 1) -> Optional[int]:
    """
    强制释放占用指定端口的所有进程。

    Args:
        port (int): 要释放的端口号
        max_try (int, optional): 最大尝试次数。值为 None，代表无限次尝试。

    Returns:
        int | None: 成功释放进程的 PID，如果不存在对应进程则返回 None。
    """

    def _kill(p: psutil.Process):
        pid: Optional[int] = None
        try_count = 0
        while max_try is None or try_count < max_try:
            if not p or not p.pid:
                return pid
            pid = p.pid
            try:
                p.kill()
                logging.info("process on port %s been killed: %s", port, pid)
                return pid
            except Exception as e:
                try_count += 1
                logging.warning(
                    "kill process on port %s failed(x%s): %s", port, try_count, e
                )
        raise RuntimeError(
            f"kill process on port {port} failed after {try_count} times"
        )

    ps = find_all_process_on_port(port)
    for p in ps:
        _kill(p)


def ensure_path(path: Union[str, Path]):
    return Path(path)
    # return path if isinstance(path, Path) else Path(path)



import math
from numbers import Number
from typing import Literal


def compute(
    operation: Literal["add", "subtract", "multiply", "divide"], *nums: Number
) -> Number:
    if not nums:
        raise ValueError("At least one number is required.")

    if len(nums) == 1:
        return nums[0]

    all_none = all(num is None for num in nums)
    if all_none:
        return None

    all_numbers = all(isinstance(num, Number) for num in nums)
    if not all_numbers:
        raise ValueError("All arguments must be numbers.")

    if operation == "add":
        return sum(nums)
    elif operation == "subtract":
        return nums[0] - sum(nums[1:])
    elif operation == "multiply":
        return math.prod(nums)
    elif operation == "divide":
        return nums[0] / math.prod(nums[1:])
    else:
        raise ValueError(f"Unknown operation: {operation}")



import logging
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler
from typing import List, Optional, Type

from pydantic import BaseModel, Field, field_validator

from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG
from logis.util.pkg_util import get_class_full_path


class LoggerDictConfig(BaseModel):
    """
    Logger配置条目
    """

    name: str
    level: int = logging.INFO
    propagate: bool = False
    handlers: List[str] = []


class HandlerDictConfig(BaseModel):
    """
    Handler配置条目
    """

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    name: str
    clazz: Type[logging.Handler]
    level: int = logging.INFO
    formatter: Optional[str] = None
    filters: List[str] = []

    filename: Optional[str] = None
    max_bytes: Optional[int] = Field(100 * 1024 * 1024, alias="maxBytes")
    backup_count: Optional[int] = Field(3, alias="backupCount")
    when: Optional[str] = "d"
    encoding: str = "utf-8"
    delay: bool = True

    def model_dump(self, by_alias=True, exclude_none=True, **kwargs):
        exclude = kwargs.pop("exclude", set())
        exclude.add("clazz")
        if self.clazz is RotatingFileHandler:
            exclude.update({"when"})
        elif self.clazz is TimedRotatingFileHandler:
            exclude.update({"max_bytes"})
        elif not issubclass(self.clazz, logging.FileHandler):
            exclude.update(
                {"filename", "max_bytes", "backup_count", "when", "encoding", "delay"}
            )
        class_path = get_class_full_path(self.clazz)
        return {
            **super().model_dump(
                by_alias=by_alias, exclude=exclude, exclude_none=exclude_none, **kwargs
            ),
            **{"class": class_path},
        }



__doc__ = """
类型模块
"""
from .base import *
from .config import *
from .echarts import *
from .exception import *
from .net import *
from .notice import *
from .page import *
from .point import *
from .spatial import *
from .unitable import *
from .updater import *



import asyncio
import inspect
from types import GeneratorType
from typing import Any, Coroutine

from logis.data_type import InvokeResult


def generate_all(v: GeneratorType):
    """
    执行生成器函数，返回所有yield值
    :param v: 生成器函数
    :return: 生成器函数返回值,包括yield出的值和函数的返回值
    """
    assert isinstance(v, GeneratorType)

    result = InvokeResult()
    result.is_generator = True
    try:
        while True:
            result.yield_values.append(next(v))
    except StopIteration as e:
        result.return_value = e.value
    return result


def invoke_async_func(coroutine: Coroutine[Any, Any, Any]):
    from logis.util.io_util import get_or_new_event_loop

    loop = get_or_new_event_loop(auto_set=True)
    return loop.run_until_complete(coroutine)


def invoke(func, *args, **kwargs):
    """
    调用函数，返回结果,支持普通函数、生成器函数

    :param func: 函数
    :param args: 位置参数
    :param kwargs: 关键字参数
    :return: 函数返回值
    """
    result = func(*args, **kwargs)
    if inspect.iscoroutinefunction(func):
        result = invoke_async_func(result)
    elif isinstance(result, GeneratorType):
        return generate_all(result)
    return InvokeResult.returns(result)



from logis.data_type import ConfigItem, QuantifiedValue, SpatialProps


def test_config_item():
    cfg = ConfigItem(label="层高", key="height", value=100, value_type="int")

    assert cfg.value_type == "int"


def test_spatial_props():
    props = SpatialProps.model_validate(
        dict(
            层高=dict(quantity=100),
            width=dict(quantity=100),
            centerPoint=dict(x=1, y=None, z=9.0),
        )
    )

    assert props.height.value == 100
    assert props.width.value == 100
    assert props.depth is None
    assert props.center_point.z == 9.0

    dc = props.model_dump()
    print(dc)

    props = SpatialProps.model_validate(dc)
    assert props.center_point.z == 9.0


def test_quantified_value():
    qv = QuantifiedValue.model_validate(dict(quantity=100, unit="m", name="高度"))
    assert qv.quantity == 100
    assert qv.unit == "m"

    dv = qv - qv

    assert isinstance(dv, QuantifiedValue)
    assert dv.quantity == 0



import logging
from typing import Any, Set, Type, Union

from pydantic import AliasChoices, AliasPath, BaseModel
from pydantic import Field as PydanticField
from sqlmodel import Field as SqlModelField


def collect_field_aliases(field: Union[PydanticField, SqlModelField]) -> Set[str]:
    """
    获取字段的别名。
    """

    def _try_collect(field: Any):
        x = getattr(field, "alias", None)
        y = getattr(field, "validation_alias", None)
        z = getattr(field, "serialization_alias", None)

        def _pick_alias(alias: Any):
            tmps = set()
            if isinstance(alias, str):
                tmps.add(alias)
            elif isinstance(alias, AliasChoices):
                for choice in alias.choices:
                    tmps.update(_pick_alias(choice))
            elif isinstance(alias, AliasPath):
                tmps.update(alias.convert_to_aliases())
            else:
                logging.warning(f"Unknown alias type: {type(alias)}")
            return tmps

        aliases = set()
        for tmp in filter(lambda x: x is not None, [x, y, z]):
            aliases |= _pick_alias(tmp)
        return aliases

    return set(_try_collect(field))


def rename_keys_by_field_name(m: Type[BaseModel], data: dict):
    """
    重命名数据字典的键，使它们与Pydantic模型的字段名称匹配。
    注意：
        1. 如果data中同时包含字段名和别名，字段名会被别名覆盖
        2. 如果有多个别名，以最后一个别名为准
    """
    for name, f in m.model_fields.items():
        # TODO: 是否可能因为集合的无序性导致不一致问题
        aliases = collect_field_aliases(f)
        for alias in aliases:
            if alias in data:
                data[name] = data.pop(alias)
    return data


def model_validate(m: Type[BaseModel], data: dict):
    """
    sqlmodel的model_validate方法特性落后于pydantic，为了解决alias问题，产生了此方法
    """
    return m.model_validate(rename_keys_by_field_name(m, data))
